#    Copyright 2013 Peter Scamardo
#    gto@dotdot.at
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
import serial
import binascii
import serial.tools.list_ports
import cmd
import sys
import time
import re
import csv
import definitions
import aldlparser
from aldlparser import hexlist2int, hexlist2str, int2hexlist

cluster_models = definitions.cluster_models
cluster_hsv_type = definitions.cluster_hsv_type
cluster_transmission = definitions.cluster_transmission
car_modules = definitions.car_modules
car_modules_hex = definitions.car_modules_hex
class VZCluster(aldlparser.ALDLParser):
    cluster_tables = {}
    cluster = car_modules_hex["Cluster"]
    bcm = car_modules_hex["BCM"]
    car_module = "cluster"
    def interpret(self, data, table):
        if table == 0x1:
            print "Software Version",data[7:10]
            print "Mfg date", data[4],data[5],"20"+data[6]
            print "Cluster Part #",hexlist2int(data[10:14])
            print "Fuel calibration part#",hexlist2int(data[16:20])
            print "Last 6 of VIN",hexlist2int(data[20:24])
            print "Cluster Serial",hexlist2int(data[24:28])

        elif table == 0x2:
            print "Odometer:",hexlist2int(data[2:6]),"MI"

        elif table == 0x4:
            print "GM Part Number:",hexlist2int(data[2:6])
            print "VAP Scratch pad:",data[6:12]
            print "Last 6 of VIN:",hexlist2int(data[16:20])
            print "Speedo Pulse perKM", hexlist2int(data[20:22])
            print "Tacho pulse per rev", hexlist2int(data[22:24])/100.
            print "Transmission", cluster_transmission[int(data[24],16)] if int(data[24],16) in range(0,4) else data[24]
            print "Startup logo", cluster_models[data[30]]
            print "Airbag configuration", data[31]
            print "HSV Serial #/Veh. year", hexlist2int(data[40:42])
            print "Cold shiftlight", int(data[43],16)/2*100,"RPM"
            print "1st gear shiftlight", int(data[44],16)/2*100,"RPM"
            print "2nd gear shiftlight", int(data[45],16)/2*100,"RPM"
            print "3rd gear shiftlight", int(data[46],16)/2*100,"RPM"
            print "4th gear shiftlight", int(data[47],16)/2*100,"RPM"
            print "5th gear shiftlight", int(data[48],16)/2*100,"RPM"
            print "Temp alarm", int(data[53],16)
            print "BCM output 0 =", int(data[54],16)
            print "BCM output 64 =", int(data[55],16)
            print "BCM output 128 =", int(data[56],16)
            print "BCM output 192 =", int(data[57],16)
            print "BCM output 255 =", int(data[58],16)
            print "Brightness% (dials):", int(data[60],16)
            print "Brightness% (pointers):", int(data[61],16)
            print "Brightness% (displays):", int(data[62],16)

        elif table == 0x6:
            print "Fuel cal part #", hexlist2int(data[40:44])
            #Mangle the ohm readings since they used 2bytes to store the ohms >:-|
            ohmdata = data[2:22]
            ohms = [int(ohmdata[i]+ohmdata[i+1],16) for i in range(0,len(ohmdata)-1,2)]
            sizedata = [int(x,16)/2. for x in data[22:32]]
            for ohmsize in zip(ohms,sizedata):
                ohm,size = ohmsize
                print "Sender %s ohms = %sL"%(ohm,size)
            print "Tank size", hexlist2int(data[32:33])/2.,'L'    
            print "Low petrol warn", hexlist2int(data[33]),'L'    
            print "Very Low petrol warn", hexlist2int(data[34]),'L'    
            print "Gauge E   = %sL"%(int(data[35],16)/2.)
            print "Gauge 1/4 = %sL"%(int(data[36],16)/2.)
            print "Gauge 1/2 = %sL"%(int(data[37],16)/2.)
            print "Gauge 3/4 = %sL"%(int(data[38],16)/2.)
            print "Gauge F   = %sL"%(int(data[39],16)/2.)
            print "Sender error low", hexlist2int(data[46:48]),'ohms'    
            print "Sender error high", hexlist2int(data[48:50]),'ohms'    
            
        elif table == 0x9:
            print "Startup Logo:",cluster_models[data[2]]
            print "HSV Serial:",hexlist2int(data[5:7])
            print "HSV Model:",cluster_hsv_type[data[7]]
            print "Cold shiftlight", int(data[8],16)/2*100,"RPM"
            print "1st gear shiftlight", int(data[9],16)/2*100,"RPM"
            print "2nd gear shiftlight", int(data[10],16)/2*100,"RPM"
            print "3rd gear shiftlight", int(data[11],16)/2*100,"RPM"
            print "4th gear shiftlight", int(data[12],16)/2*100,"RPM"
            print "5th gear shiftlight", int(data[13],16)/2*100,"RPM"
            print "Shift light on:", True if int(data[15],16) & 0x10 else False
            print "Custom text:>", "".join(data[17:27]).decode("hex"),"<"
    def unlock(self):
        '''If in car, make sure to send no chat before using this function'''

        #####
        ##Thanks to beerman and 06phantom for figuring out this algorithm
        ####
        message = [ 0x11, 0x00 ]
        response = self.send_message(self.cluster, message)
        if len(response) == 0:
            print "No response from cluster"
        else:
            seed = int(response[2] + response[1],16)
            print "Seed:", hex(seed),
            key = abs(((seed-0xBE86)%0xFFFF) - 0xFFFF)
            if seed > 0xBE86:
                print "seed > BE86",
                key = abs(~key)
            print "Key:", hex(key)
            unlox = int2hexlist(key)
            unlox = [ 0x11 , 0x01 ] + unlox
            print unlox
            unloxresponse = self.send_message(self.cluster, unlox)
            print unloxresponse
            if int(unloxresponse[1]+unloxresponse[2],16) == key:
                print "Cluster Unlocked!"

    def do_unlock(self, line):
        '''unlock
        This command unlocks the cluster for writing directly to memory 
        addresses.  It takes no arguments. '''
        if self.leetmode:
            try:
                #disable bcm chatter##
                incarstate = self.incar 
                if incarstate == True:
                    self.nochat(self.bcm)
                self.incar = False
                #####
                self.unlock()
            finally: 
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate
        else:
            print "1337 mode not enabled."

    def do_show(self, table):
        '''show [table_num]
        This shows a table that has already been read into memory'''
        print self.cluster_tables[int(table)]
        self.interpret(self.cluster_tables[int(table)], int(table))
    def read(self, table):
        hextable = int(table)
        message = [  0x01, hextable ]
        
        #create and write message
        retry = 1
        while retry <= 5:
            response = self.send_message(self.cluster, message)
            if len(response) > 2:
                #save message in cluster_tables
                self.cluster_tables[hextable] = response
                return response
            print "retry #%s"%retry
            retry += 1
        
    def do_read(self, table):
        '''read [table_num]
        This reads the data tables from the cluster.  Valid tables are 0-9.
        '''
        response = self.read(table)
        print response
        #Print interesting things about the table
        if len(response) >= 2: 
            aldl_mode = int(response[0],16)
            print "ALDL Mode:", hex(aldl_mode),
            table = int(response[1],16)
            print "(read configuration table)",
            print "Table:", hex(table)
            self.interpret(self.cluster_tables[table], table)

    def readmem(self, memloc):
        memloc = int2hexlist(int(memloc,16))
        message = [ 0x22 ] + memloc
        #send message
        response = self.send_message(self.cluster, message)
        return response
    def do_readmem(self, memloc):
        '''readmem [memory_address]
        This reads an arbitrary 2 byte memory location (max tested on GTO 
        cluster is 137), input format is HEX with no prefixes. 
        The program will attempt to interpret the data where possible.
        '''
        if re.search("[^0-9a-fA-F ]",memloc):
            print "Invalid hex character for memory location"
        elif self.leetmode:
            response = self.readmem(memloc)
            print "Response:", response
            if response[1] == "08":
                print "Cluster is not unlocked!"
            elif response[1] == "00":
                self.interpretmem(response)
        else:
            print "1337 mode not enabled."

    def writemem(self, addr, data):
        read_response = self.readmem(addr)
        if len(read_response) > 4:
            #check to make sure the data being written is the same length
            memloc = int2hexlist(int(addr,16))
            datalen = len(read_response[4:])
            print "Data length is",hex(datalen)
            parsedata = [int(x.strip(),16) for x in data.split(",")]
            if len(parsedata) == datalen:
                message = [ 0x20, 0x55 ] + memloc + parsedata
                #print "would have written", message
                print "Writing to", memloc
                response = self.send_message(self.cluster, message)
                print response
                self.checkwrite()
            else:
                print "Data is not the same length as in cluster"
                print "Yours:", parsedata
                print "Cluster:", response[4:]
        else:
            print "Read request did not return a valid message.", memloc

    def do_writemem(self, line):
        '''writemem [mem_address] = [byte1],[byte2],...
        This allows you to write arbitrary data to memory, the 
        cluster must be unlocked with the unlock command to do this.
        Be careful!  If you brick something this program can't help you'''
        if self.leetmode:
            input = line.split("=")
            if len(input) != 2:
                print "Invalid number of arguments, format should be:"
                print "<mem_address> = <data>,<to>,<write>"
            else:
                carstate = self.incar
                #disable bcm chatter##
                incarstate = self.incar 
                if incarstate == True:
                    self.nochat(self.bcm)
                    time.sleep(.25)
                self.incar = False
                #####

                addr = input[0].strip()
                data = input[1].strip()
                self.writemem(addr, data)
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate
        else:
            print "1337 mode not enabled."
    def checkwrite(self):
        '''This sends a mode 21 request to see if the cluster
        is done writing the mode 20 request'''
        message = [ 0x21 , 0x00 ] 
        #send message
        for i in range(5):
            response = self.send_message(self.cluster, message)
            print "Response:", response
            if len(response) == 0:
                print "no response!"
            if response[1] == "04":
                print "Waiting to complete"
            elif response[1] == "00":
                print "Write complete"
                break
            time.sleep(1)
    def do_preset(self, line):
        '''preset [preset-name]
        This will load and execute a preset command to change something
        in the cluster. 

        Available presets are:
        shiftlight-on
        shiftlight-off
        drl-stock
        drl-lightoff
        drl-fulloff
        mph-dim
        temp-gauge
        temp-gauge-stock
        overtemp=[number] (100-150*C)
        startup-logo=[0-5]
        hsv-number=[number] (0-9999)
        custom-text=[text]
        rpmset=[coldRPM],[1st],[2nd],[3rd],[4th],[5th]
        transmission=[0-3]
        speedometer=[PulsePerKM]
        '''
        if line.startswith("startup-logo"):
            incarstate = self.incar 
            try:
                data = line.split("=")
                if len(data) != 2:
                    print "Wrong data format, no equal sign"
                elif re.search("[^0-9 ]",data[1]):
                    print "Invalid characters entered"
                    badchar = re.search("[^0-9 ]",data[1]).start()
                    print data[1]
                    print " "*badchar+"^"
                elif int(data[1],16) not in range(0,6):
                    print "startup logo value must be between 0 and 5"
                else:
                    #disable bcm chatter##
                    incarstate = self.incar 
                    if incarstate == True:
                        self.nochat(self.bcm)
                    self.incar = False
                    #####
                    print "===Current Config==="
                    self.read("9")
                    self.onecmd("show 9")
                    #Set shiftlight to on
                    self.cluster_tables[9][2] = str(int(data[1],16)).zfill(2)
                    print "===New Config==="
                    self.onecmd("show 9")
                    print "===Write config==="
                    self.onecmd("writetable 9=%(table)s"%{"table": ",".join(self.cluster_tables[9][2:])} )
            finally:
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate

        elif line == "shiftlight-on":
            try:
                #disable bcm chatter##
                incarstate = self.incar 
                if incarstate == True:
                    self.nochat(self.bcm)
                self.incar = False
                #####
                print "===Current Config==="
                self.read("9")
                self.onecmd("show 9")
                #Set shiftlight to on
                self.cluster_tables[9][15] = "90"
                print "===New Config==="
                self.onecmd("show 9")
                print "===Write config==="
                self.onecmd("writetable 9=%(table)s"%{"table": ",".join(self.cluster_tables[9][2:])} )
            finally:
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate

        elif line == "shiftlight-off":
            try:
                #disable bcm chatter##
                incarstate = self.incar 
                if incarstate == True:
                    self.nochat(self.bcm)
                self.incar = False
                #####
                print "===Current Config==="
                self.read("9")
                self.onecmd("show 9")
                #Set shiftlight to on
                self.cluster_tables[9][15] = "80"
                print "===New Config==="
                self.onecmd("show 9")
                print "===Write config==="
                self.onecmd("writetable 9=%(table)s"%{"table": ",".join(self.cluster_tables[9][2:])} )
            finally:
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate

        elif line == "drl-stock":
            try:
                #disable bcm chatter##
                incarstate = self.incar 
                if incarstate == True:
                    self.nochat(self.bcm)
                self.incar = False
                #####
                self.unlock()
                print "===Writing New Config==="
                self.writemem("37", "C0")
            finally:
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate


        elif line == "drl-lightoff":
            try:
                #disable bcm chatter##
                incarstate = self.incar 
                if incarstate == True:
                    self.nochat(self.bcm)
                self.incar = False
                #####
                self.unlock()
                print "===Writing New Config==="
                self.writemem("37", "80")
            finally:
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate


        elif line == "drl-fulloff":
            try:
                #disable bcm chatter##
                incarstate = self.incar 
                if incarstate == True:
                    self.nochat(self.bcm)
                self.incar = False
                #####
                self.unlock()
                print "===Writing New Config==="
                self.writemem("37", "00")
            finally:
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate


        elif line == "mph-dim":
            try:
                #disable bcm chatter##
                incarstate = self.incar 
                if incarstate == True:
                    self.nochat(self.bcm)
                self.incar = False
                #####
                self.unlock()
                print "===Writing New Config==="
                self.writemem("33", "28,18")
            finally:
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate


        elif line.startswith("hsv-number"):
            incarstate = self.incar 
            try:
                data = line.split("=")
                if len(data) != 2:
                    print "Wrong data format, no equal sign"
                elif re.search("[^0-9 ]",data[1]):
                    print "Invalid characters entered"
                    badchar = re.search("[^0-9 ]",data[1]).start()
                    print data[1]
                    print " "*badchar+"^"
                elif int(data[1]) > 9999:
                    print "Number too large. Max is 9999."
                else:
                    #disable bcm chatter##
                    incarstate = self.incar 
                    if incarstate == True:
                        self.nochat(self.bcm)
                    self.incar = False
                    #####
                    print "===Current Config==="
                    self.read("9")
                    self.onecmd("show 9")
                    #Set hsv number
                    hsvnum = hex(int(data[1],10))[2:].zfill(4)
                    self.cluster_tables[9][5] = hsvnum[:2]
                    self.cluster_tables[9][6] = hsvnum[2:]
                    print "===New Config==="
                    self.onecmd("show 9")
                    print "===Write config==="
                    self.onecmd("writetable 9=%(table)s"%{"table": ",".join(self.cluster_tables[9][2:])} )
            finally:
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate

        elif line.startswith("custom-text"):
            incarstate = self.incar 
            try:
                data = line.split("=")
                if len(data) != 2:
                    print "Wrong data format, no equal sign"
                elif re.search("[^0-9a-zA-Z ]",data[1]):
                    print "Invalid characters entered"
                    badchar = re.search("[^0-9a-zA-Z]",data[1]).start()
                    print data[1]
                    print " "*badchar+"^"
                elif len(data[1]) > 10:
                    print "Too many characters. Max is 10."
                else:
                    #disable bcm chatter##
                    incarstate = self.incar 
                    if incarstate == True:
                        self.nochat(self.bcm)
                    self.incar = False
                    #####
                    print "===Current Config==="
                    self.read("9")
                    self.onecmd("show 9")
                    #Set custom text
                    self.cluster_tables[9][7] = "0D"
                    #Set startup logo to HSV
                    self.cluster_tables[9][2] = "01"
                    #Zero custom text field
                    txt_offset = 17
                    self.cluster_tables[9] = self.cluster_tables[9][:17] + [ '00' ] *10 
                    for char in data[1]:
                        self.cluster_tables[9][txt_offset] = hex(ord(char))[2:]
                        txt_offset += 1 
                    print "===New Config==="
                    self.onecmd("show 9")
                    print "===Write config==="
                    self.onecmd("writetable 9=%(table)s"%{"table": ",".join(self.cluster_tables[9][2:])} )
            finally:
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate


        elif line.startswith("rpmset"):
            incarstate = self.incar 
            try:
                data = line.split("=")
                if len(data) != 2:
                    print "Wrong data format, no equal sign"
                elif re.search("[^0-9, ]",data[1]):
                    print "Invalid characters entered"
                    badchar = re.search("[^0-9, ]",data[1]).start()
                    print data[1]
                    print " "*badchar+"^"
                elif len(data[1].split(",")) != 6:
                    print "Incorrect # of shift points, need 6. Cold,1st,2nd,3rd,4th,5th"
                else:
                    #disable bcm chatter##
                    if incarstate == True:
                        self.nochat(self.bcm)
                    self.incar = False
                    #####
                    self.read("9")
                    print "Read table 9"
                    print "===Current Config==="
                    self.onecmd("show 9")
                    #Set custom text
                    rpm = data[1].split(",")
                    self.cluster_tables[9][8] = hex(int(rpm[0])/100*2)[2:]
                    self.cluster_tables[9][9] = hex(int(rpm[1])/100*2)[2:]
                    self.cluster_tables[9][10] = hex(int(rpm[2])/100*2)[2:]
                    self.cluster_tables[9][11] = hex(int(rpm[3])/100*2)[2:]
                    self.cluster_tables[9][12] = hex(int(rpm[4])/100*2)[2:]
                    self.cluster_tables[9][13] = hex(int(rpm[5])/100*2)[2:]
                    print "===New Config==="
                    self.onecmd("show 9")
                    print "===Write config==="
                    self.onecmd("writetable 9=%(table)s"%{"table": ",".join(self.cluster_tables[9][2:])} )
            finally:
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate


        elif line.startswith("transmission"):
            try:
                incarstate = self.incar 
                data = line.split("=")
                if len(data) != 2:
                    print "Wrong data format, no equal sign"
                elif re.search("[^0-9, ]",data[1]):
                    print "Invalid characters entered"
                    badchar = re.search("[^0-9, ]",data[1]).start()
                    print data[1]
                    print " "*badchar+"^"
                elif int(data[1],16) not in range(0,4):
                    print "Transmission value must be between 0 and 3"
                else:
                    #disable bcm chatter##
                    incarstate = self.incar 
                    if incarstate == True:
                        self.nochat(self.bcm)
                    self.incar = False
                    #####
                    
                    self.read("4")
                    print "Read table 4"
                    print "===Current Config==="
                    self.onecmd("show 4")
                    self.cluster_tables[4][24] = hex(int(data[1],16))[2:].zfill(2)
                    print "===New Config==="
                    self.onecmd("show 4")
                    print "===Write config==="
                    self.onecmd("writetable 4=%(table)s"%{"table": ",".join(self.cluster_tables[4][2:])} )
            finally:
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate
                #####
        elif line == "temp-gauge":
            try:
                #disable bcm chatter##
                incarstate = self.incar 
                if incarstate == True:
                    self.nochat(self.bcm)
                self.incar = False
                #####
                self.unlock()
                print "===Writing New Config==="
                #Temp gauge calibration (no more idiot gauge!)
                self.writemem("3e", "3C,4B,50,52,5D,63,69,6B,76")
                #Overheat alarm (220*F)
                #self.writemem("47", "6E")
            finally:
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate
        elif line == "temp-gauge-stock":
            try:
                #disable bcm chatter##
                incarstate = self.incar 
                if incarstate == True:
                    self.nochat(self.bcm)
                self.incar = False
                #####
                self.unlock()
                print "===Writing New Config==="
                #Temp gauge calibration
                self.writemem("3e", "3C,50,55,6E,71,74,77,79,7C")
                #Overheat alarm (248*F)
                self.writemem("47", "78")
            finally:
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate
        elif line.startswith("overtemp"):
            try:
                incarstate = self.incar 
                data = line.split("=")
                if len(data) != 2:
                    print "Wrong data format, no equal sign"
                elif re.search("[^0-9]",data[1]):
                    print "Invalid characters entered"
                    badchar = re.search("[^0-9]",data[1]).start()
                    print data[1]
                    print " "*badchar+"^"
                elif int(data[1],10) not in range(100,150):
                    print "Overtemp warning must be an integer between 100*C and 150*C"
                else:
                    #disable bcm chatter##
                    incarstate = self.incar 
                    if incarstate == True:
                        self.nochat(self.bcm)
                    self.incar = False
                    #####
                    self.unlock()
                    print "===Writing New Config==="
                    tempvalue = hex(int(data[1],10))
                    self.writemem("47", tempvalue)
            finally:
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate
                #####

        elif line.startswith("speedometer"):
            try:
                incarstate = self.incar 
                data = line.split("=")
                if len(data) != 2:
                    print "Wrong data format, no equal sign"
                elif re.search("[^0-9 ]",data[1]):
                    print "Invalid characters entered"
                    badchar = re.search("[^0-9 ]",data[1]).start()
                    print data[1]
                    print " "*badchar+"^"
                else:
                    #disable bcm chatter##
                    incarstate = self.incar 
                    if incarstate == True:
                        self.nochat(self.bcm)
                    self.incar = False
                    #####
                    
                    self.read("4")
                    print "Read table 4"
                    print "===Current Config==="
                    self.onecmd("show 4")
                    ppk = hex(int(data[1],10))[2:].zfill(4)
                    self.cluster_tables[4][20] = ppk[:2]
                    self.cluster_tables[4][21] = ppk[2:]
                    print "===New Config==="
                    self.onecmd("show 4")
                    print "===Write config==="
                    self.onecmd("writetable 4=%(table)s"%{"table": ",".join(self.cluster_tables[4][2:])} )
            finally:
                #Re-enable bcm chatter
                if incarstate == True:
                    self.chat(self.bcm)
                self.incar = incarstate
                #####
                
        else:
            print "Preset %s not found"%line
    def do_writetable(self, line):
        '''writetable [table_number] = [byte1],[byte2]...
        This allows you to write a string of values to the specified table.
        '''
        input = line.split("=")
        if len(input) != 2:
            print "Invalid number of arguments, format should be:"
            print "<table_number> = <table_contents>,<data>,<to>,<write>"
        elif re.search("[^a-fA-F0-9, ]", input[1]):
            print "Invalid character found:"
            badchar = re.search("[^a-fA-F0-9, ]", input[1]).start()
            print input[1],"\n"+" "*badchar+"^"
        else:
            table = int(input[0],16)
            data = [int(x.strip(),16) for x in input[1].split(",")]
            message = [0x28, table] + data
            #print message
            response = self.send_message(self.cluster, message)
            print response

    def do_readmemALL(self, line):
        '''readmemALL
        Reads all the memory in range 0x01-0x137. Useful for hackers
        '''
        if self.incar == True:
            print "I wouldn't recommend doing a full-read in car."
        elif self.leetmode:
            if line != "":
                outputfile = open(line,"w")
                csvoutput = csv.writer(outputfile, delimiter=',', dialect='excel')
            nextaddress = 0x01
            try:
                self.serialport.open()
                for address in range(0x01,0x138):
                    memloc = int2hexlist(address)
                    message = [ 0x22 ] + memloc

                    response = self.send_message_quiet(self.cluster, message)

                    print hex(address),"Data:", response[4:]
                    if line != "":
                        #outputfile.write(hex(address)+" Data: "+str(response[4:]) +"\r\n")
                        csvoutput.writerow([hex(address)] + response[4:])
                    nextaddress = address + len(response[4:])
            finally:
                self.serialport.close()
                
            if line != "":
                outputfile.close()
        else:
            print "1337 mode not enabled."
        
    def do_readALL(self, line):
        if line != "":
            outputfile = open(line,"w")
            csvoutput = csv.writer(outputfile, delimiter=',', dialect='excel')
        try:
            #disable bcm chatter##
            incarstate = self.incar 
            if incarstate == True:
                self.nochat(self.bcm)
            self.incar = False
            #####
            for tbl in range(0,10):
                print self.read(tbl)
                if line != "":
                    csvoutput.writerow(["Table " + hex(tbl)] + self.cluster_tables[tbl][2:])
        finally:
            #Re-enable bcm chatter
            if incarstate == True:
                self.chat(self.bcm)
            self.incar = incarstate

        if line != "":
            outputfile.close()
    def interpretmem(self,data):
        import definitions
        memory_defs = definitions.cluster_memory_definitions
        #print memory_defs.keys()
        #data = data
        #print data
        #print int(data[5],16)
        #print data[6:]
        try:
            if len(data) < 4:
                print "short",
                print data
                return
            addr = int(data[2]+data[3],16)
            print hex(addr),
            print memory_defs[addr](data[4:])
            print "Next Mem address::",hex(addr+len(data[4:]))
            print "++++++++++++++++++++++\n"
        except KeyError as e:
            print "Address:",hex(e.message),"not found in memory_defs"
            print "Mode/Addr:",data[:1], data[2]+data[3]
            if len(data[4:]) == 0:
                print "No data returned"
                print "-----------------------"
            else:
                print "Hex data:", data[4:]
                print "\tByte as int data:", [int(x,16) for x in data[4:]]
                print "\tAll data as int:", hexlist2int(data[4:])
                print "\tAll data as str:", repr( hexlist2str(data[4:]))
                print "Next Mem address::", hex(e.message+len(data[4:]))
                print "-----------------------"
    def do_1337vending(self, line):
        self.onecmd("open com4")
        self.onecmd("iddqd")
        self.onecmd("incar")
